/*
 * Copyright 2002 Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.io.Serializable;

/**
 * The game arena that holds all the spatial objects.
 *
 * @author Henri Sivonen, hsivonen@iki.fi
 */
public final class Arena implements Serializable {
    /**
     * Constant for the height of the arena. Must match
     * <code>SpaceArea.WINDOW_HEIGHT</code>.
     */
    public static final double HEIGHT = 480.0;

    /**
     * Constant for the width of the arena. Must match
     * <code>SpaceArea.WINDOW_WIDTH</code>.
     */    
    public static final double WIDTH = 640.0;
    
    /**
     * An array holding all the spaceships
     */
    private final SpaceShip[] spaceShips;

    /**
     * An array holding all the asteroids
     */
    private final Asteroid[] asteroids;

    /**
     * The starting time of the level.
     */
    private final long levelStarts;

    /**
     * Return an array holding references to all
     * spaceships. The caller should not write to the
     * array.
     *
     * @return an array holding references to all
     *         spaceships
     */
    public final SpaceShip[] getSpaceShips() {
        return this.spaceShips;
    }

    /**
     * Return an array holding references to all
     * asteroids. The caller should not write to the
     * array.
     *
     * @return an array holding references to all
     *         asteroids
     */   
    public final Asteroid[] getAsteroids() {
        return this.asteroids;
    }

    /**
     * Returns the starting time of the level.
     *
     * @return the starting time of the level
     */
    public final long getLevelStarts() {
        return this.levelStarts;
    }

    /**
     * Contructs an <code>Arena</code> including the
     * <code>SpatialObject</code>s in it for a new level.
     *
     * @param levelNumber the level number
     * @param timestamp the starting time of the level
     */
    public Arena(int levelNumber, long timestamp) {
        this.levelStarts = timestamp;
        int numOfAsteroids = levelNumber + 5;
        double asteroidVelocity = 0.02 + levelNumber * 0.005;
        this.asteroids = new Asteroid[numOfAsteroids * 7];
        this.spaceShips = new SpaceShip[2];
        this.spaceShips[0] = new SpaceShip((WIDTH / 2.0) + 30.0,
                                           HEIGHT / 2.0,
                                           timestamp);
        this.spaceShips[1] = new SpaceShip((WIDTH / 2.0) - 30.0,
                                           HEIGHT / 2.0,
                                           timestamp);
        double angle, velocity, velocityX, velocityY, locationX, locationY;
        // the asteroid table layout is documented in Asteroid.java
        for(int i = 0; i < numOfAsteroids * 7; i += 7) {
            angle = 2 * Math.PI * Math.random();
            velocity = asteroidVelocity * (0.9 + 0.2 * Math.random());
            velocityX = velocity * Math.cos(angle);
            velocityY = velocity * Math.sin(angle);
            if(Math.random() < HEIGHT / (HEIGHT + WIDTH)) {
                locationX = 0.0;
                locationY = HEIGHT * Math.random();
            } else {
                locationX = WIDTH * Math.random();
                locationY = 0.0;
            }
            this.asteroids[i] = new Asteroid(locationX,
                                             locationY,
                                             velocityX,
                                             velocityY,
                                             timestamp,
                                             Asteroid.LARGE);
        }
        Asteroid.generateResidualAsteroids(this.asteroids);
    }
}
