/*
 * Copyright 2002 Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.awt.*;
import java.awt.geom.*;
import java.io.IOException;

/**
 * The representation of an asteroid in the model.
 *
 * <h3><a name="asteroid-table-layout"></a>The Layout of the
 * Asteroid Table</h3>
 * <p>The length of the asteroid table is required to be a multiple
 * of 7. The first large asteroid is at index <code>0</code>. The
 * second is at <code>7</code>, etc.</p>
 * <p>When a large asteroid is at index <code>i</code>, its residual
 * medium-sized asteroids are at indeces <code>i + 1</code> and
 * <code>i + 4</code>.</p>
 * <p>When a medium-sized asteroid is at index <code>i</code>, its
 * residual small asteroids are at indeces <code>i + 1</code> and
 * <code>i + 2</code>.</p>
 *
 * <h3>A Note About Thread Safety</h3>
 * <p>The setters of an <code>Asteroid</code> may be used from
 * multiple threads. Also, the getters may be used from
 * a single thread safely even when there are other
 * thread using the setters. However, there may
 * <strong>not</strong> be multiple threads calling
 * <code>getTransformedShape</code>. It is designed
 * to be used from a single thread only.</p>
 * <p>Also, that thread should not retain references to
 * objects returned by <code>getTransformedShape</code>
 * across the method calls. When the method is called
 * the object it returned previously will no longer be valid.
 * </p>
 *
 * @author Henri Sivonen, hsivonen@iki.fi
 */
public class Asteroid extends SpatialObject {
    /**
     * Constant for indicating that the asteroid is small
     */
    public static final int SMALL = 0;

    /**
     * Constant for indicating that the asteroid is medium-sized
     */
    public static final int MEDIUM = 1;

    /**
     * Constant for indicating that the asteroid is large
     */
    public static final int LARGE = 2;

    /**
     * Constant used when modeling the extra energy released when
     * a large asteroid explodes.
     */   
    public static final double LARGE_EXPLOSION_EXTRA = 0.02;

    /**
     * Constant used when modeling the extra energy released when
     * a medium-sized asteroid explodes.
     */
    public static final double MEDIUM_EXPLOSION_EXTRA = 0.015;

    /**
     * Constant used when calculating the movement of asteroid debris.
     */
    public static final float DEBRIS_VELOCITY_FACTOR = 0.002f;

    /**
     * The lifetime of asteroid debris.
     */
    public static final float DEBRIS_LIFETIME = 3000.0f;

    /**
     * The amount of time used for one outline color in an explosion.
     */
    public static final float COLOR_STEP =
        DEBRIS_LIFETIME / SpatialObject.COLOR_TABLE.length;

    /** 
     * the outline color as an index to 
     * <code>SpatialObject.COLOR_TABLE</code> 
     */
    private transient int outlineColor;

    /**
     * Generates residual asteroids in the array of asteroids
     * passed as the argument.
     *
     * @param asteroids the <a href"#asteroid-table-layout">asteroid table</a>
     */
    public static void generateResidualAsteroids(Asteroid[] asteroids) {
        double extra, baseExtra, angle;
        int residual1, residual2;
        for(int i = 0; i < asteroids.length; i++) {
            if(asteroids[i].size == LARGE) {
                baseExtra = LARGE_EXPLOSION_EXTRA;
                residual1 = i + 1;
                residual2 = i + 4;
            } else if(asteroids[i].size == MEDIUM) {
                baseExtra = MEDIUM_EXPLOSION_EXTRA;
                residual1 = i + 1;
                residual2 = i + 2;
            } else {
                continue;
            }
            angle = 2 * Math.PI * Math.random();
            extra = baseExtra * (0.75 + 0.5 * Math.random());
            asteroids[residual1] = new Asteroid(0.0,
                                                0.0,
                                                extra
                                                * Math.cos(angle)
                                                + 0.5
                                                * asteroids[i].velocityX,
                                                extra
                                                * Math.sin(angle)
                                                + 0.5
                                                * asteroids[i].velocityY,
                                                -1L,
                                                asteroids[i].size - 1);
            asteroids[residual2] = new Asteroid(0.0,
                                                0.0,
                                                -extra
                                                * Math.cos(angle)
                                                + 0.5
                                                * asteroids[i].velocityX,
                                                -extra
                                                * Math.sin(angle)
                                                + 0.5
                                                * asteroids[i].velocityY,
                                                -1L,
                                                asteroids[i].size - 1);
        }
    }

    /**
     * Makes an asteroid explode and initalizes the residual asteroids
     * accordingly.
     *
     * @param asteroids the <a href"#asteroid-table-layout">asteroid table</a>
     * @param index the index of the asteroid to be blown up
     * @param locationX the x coordinate of the asteroid at the time of the
     *                  explosion
     * @param locationY the y coordinate of the asteroid at the time of the
     *                  explosion
     * @param timestamp the point in time when the explosion happens
     * @param onServer  a boolean indicating whether this method is called
     *                  from the server thread
     */
    public static void blowUpAsteroid(Asteroid[] asteroids,
                                      int index,
                                      double locationX,
                                      double locationY,
                                      long timestamp,
                                      boolean onServer)
    {
        Asteroid explodingAsteroid = asteroids[index];
        int residual1, residual2;
        if(explodingAsteroid.getSize() != Asteroid.SMALL) {
            if(explodingAsteroid.getSize() == Asteroid.LARGE) {
                residual1 = index + 1;
                residual2 = index + 4;
            } else {
                residual1 = index + 1;
                residual2 = index + 2;
            }
            asteroids[residual1].setLocationAndAngle(locationX,
                                                     locationY,
                                                     0.0,
                                                     timestamp);
            asteroids[residual2].setLocationAndAngle(locationX,
                                                     locationY,
                                                     0.0,
                                                     timestamp);
            asteroids[residual1].makeVisible();
            asteroids[residual2].makeVisible();
        }
        if(onServer) {
            explodingAsteroid.explodeOnServer();
        } else {
            explodingAsteroid.explode(timestamp);
        }
    }
    
    /**
     * The angular velocity (radians per millisecond) of the asteroid.
     */
    private double angularVelocity;
   
    /**
     * The size of the asteroid. <code>SMALL</code>, <code>MEDIUM</code> or
     * <code>LARGE</code>
     */
    private int size;

    /**
     * A boolean indicating whether the asteroid is visible.
     */
    private boolean visible;

    /**
     * A boolean indicating whether the asteroid is currently exploding
     * (that is, whether debris is visible).
     */
    private boolean exploding;

    /**
     * The rotation angle (in radians) of the spatial object at the
     * moment indicated by <code>timestamp</code>.
     */
    private double angle;

    /**
     * The number of hits it takes to make this asteroid explode.
     */
    private transient int hitsToExplosion;

    /**
     * An array for storing the coordinates of the vertices of this asteroid.
     * The array is in the format accepted by <code>AffineTransform</code>:
     * x0, y0, ... xn, yn.
     */
    private float[] vertices;

    private transient int points;

    /**
     * An array for storing the coordinates of the centers of the edges
     * of this asteroid.
     * The array is in the format accepted by <code>AffineTransform</code>:
     * x0, y0, ... xn, yn.
     */
    private transient float[] centers;
    
    /**
     * Contructs a new asteroid. The geometry is created automatically.
     * The angular velocity is chosen automatically. Large asteroids
     * are visible initially. Other asteroids are invisible by default.
     *
     * @param locationX the x coordinate of the initial location
     * @param locationY the y coordinate of the initial location
     * @param velocityX the x component of the velocity
     * @param velocityY the y component of the velocity
     * @param timestamp the moment in game (server) time for which the
     *                  location is valid
     * @param size the size of the asteroid: <code>SMALL</code>,
     *             <code>MEDIUM</code> or <code>LARGE</code>
     */
    public Asteroid(double locationX,
                    double locationY,
                    double velocityX,
                    double velocityY,
                    long timestamp,
                    int size)
    {
        super(locationX, locationY, velocityX, velocityY, timestamp);
        this.angularVelocity = 0.0005 + 0.0005 * Math.random();
        
        this.size = size;
        this.angle = 0.0;
        this.transform = new AffineTransform();
        double referenceRadius;
        int numOfVertices;
        switch(size) {
            case SMALL:
                referenceRadius = 10.0;
                numOfVertices = 7;
                this.visible = false;
                this.hitsToExplosion = 1;
                this.points = 5;
                break;
            case MEDIUM:
                referenceRadius = 20.0;
                numOfVertices = 8;
                this.visible = false;
                this.hitsToExplosion = 2;
                this.points = 10;
                break;
            case LARGE:
                referenceRadius = 30.0;
                numOfVertices = 9;
                this.visible = true;
                this.hitsToExplosion = 3;
                this.points = 20;
                break;
            default:
                throw new IllegalArgumentException();
        }
        /*
         * This is run only between levels on the server.
         * No need to optimize the math further.
         */
        this.transform = new AffineTransform();
        this.geometry = new GeneralPath(GeneralPath.WIND_NON_ZERO,
                                        numOfVertices + 2);
        double angleIncrement = 2.0 * Math.PI / numOfVertices;
        double angle;
        double radius;
        float xCoord;
        float yCoord;
        this.vertices = new float[numOfVertices * 2];
        for(int i = 0; i < numOfVertices; i++) {
            angle = i * angleIncrement;
            angle += Math.random() * angleIncrement / 2.0 - angleIncrement / 4.0;
            radius = referenceRadius * (0.7 + 0.6 * Math.random());
            xCoord = (float)(radius * Math.cos(angle));
            yCoord = (float)(radius * Math.sin(angle));
            this.vertices[i * 2] = xCoord;
            this.vertices[i * 2 + 1] = yCoord;
            if(i == 0) {
                this.geometry.moveTo(xCoord, yCoord);
            } else {
                this.geometry.lineTo(xCoord, yCoord);
            }
        }
        this.geometry.closePath();
    }

    /**
     * Returns a properly transformed <code>Shape</code> representing
     * the spatial object at <code>time</code>
     *
     * @param time the moment in game (server) time for which the
     *             representation is requested
     * @return the <code>Shape</code> representation at <code>time</code>
     *         or null if the object isn't visible
     */
    public Shape getTransformedShape(long time) {
        if(this.visible) {
            this.outlineColor = SpatialObject.COLOR_TABLE_WHITE;
            double deltaT, theta, x, y, cosTheta, sinTheta;
            synchronized(this) {
                deltaT = (double)(time - this.timestamp);
                theta = this.angle + this.angularVelocity * deltaT;
                x = this.locationX + this.velocityX * deltaT;
                y = this.locationY + this.velocityY * deltaT;
            }
            x = moveToRangeX(x);
            y = moveToRangeY(y);
            cosTheta = Math.cos(theta);
            sinTheta = Math.sin(theta);
            this.transform.setTransform(cosTheta,
                                        sinTheta,
                                        -sinTheta,
                                        cosTheta,
                                        x,
                                        y);
            return this.geometry.createTransformedShape(this.transform);
        } else if(this.exploding) {
            // no need to synchronize this explosion stuff since
            // the server has abandoned this object
            double deltaT, theta;
            float cosTheta, sinTheta, x, y, startX, startY, endX, endY,
                  centerX, centerY, rStartX, rEndX, rStartY, rEndY;
            deltaT = (double)(time - this.timestamp);
            if(deltaT > DEBRIS_LIFETIME) {
                this.exploding = false;
                return null;
            }

            // calculate the color from the time delta
            this.outlineColor =  16 - (int)(deltaT / COLOR_STEP);
            if(this.outlineColor < 0) {
                this.outlineColor = 0;
            }
            if(this.outlineColor > 16) {
                this.outlineColor = 16;
            }

            x = (float)this.locationX;
            y = (float)this.locationY;
            int numOfVertices = this.centers.length / 2;
            GeneralPath debrisGeometry = new GeneralPath(GeneralPath.WIND_NON_ZERO,
                                        this.centers.length + 2);
            float deltaTf = (float)deltaT;
            theta = this.angle + this.angularVelocity * deltaT;
            cosTheta = (float)Math.cos(theta);
            sinTheta = (float)Math.sin(theta);
            int j;
            for(int i = 0; i < numOfVertices - 1; i++) {
                j = i + 1;
                if(j == numOfVertices) {
                    j = 0;
                }
                startX = this.vertices[i * 2];
                startY = this.vertices[i * 2 + 1];
                endX = this.vertices[j * 2];
                endY = this.vertices[j * 2 + 1];
                centerX = this.vertices[i * 2];
                centerY = this.vertices[i * 2 + 1];

                // translate to origin
                startX -= centerX;
                endX -= centerX;
                startY -= centerY;
                endY -= centerY;

                // rotate
                rStartX = startX * cosTheta - startY * sinTheta;
                rEndX = endX * cosTheta - endY * sinTheta;
                rStartY = startX * sinTheta + startY * cosTheta;
                rEndY = endX * sinTheta + endY * cosTheta;

                // translate
                float deltaX = DEBRIS_VELOCITY_FACTOR * centerX * deltaTf + x + centerX;
                float deltaY = DEBRIS_VELOCITY_FACTOR * centerY * deltaTf + y + centerY;
                startX = rStartX + deltaX;
                endX = rEndX + deltaX;
                startY = rStartY + deltaY;
                endY = rEndY + deltaY;

                // draw
                debrisGeometry.moveTo(startX, startY);
                debrisGeometry.lineTo(endX, endY);
            }
            return debrisGeometry;
        } else {
            return null;
        }
    }

    /**
     * Sets the location and angle of this asteroid.
     * This method is supposed to be called from
     * <code>ClientModelController</code>.
     *
     * @param locationX the x coordinate of the new location
     * @param locationY the y coordinate of the new location
     * @param angle the new angle
     * @param timestamp the moment in game (server) time for which
     *                  at which the other values were valid.
     */
    public void setLocationAndAngle(double locationX,
                                    double locationY,
                                    double angle,
                                    long timestamp)
    {
        synchronized(this) {
            this.locationX = locationX;
            this.locationY = locationY;
            this.angle = angle;
            this.timestamp = timestamp;
        }
    }

    /**
     * Updates the location and angle of this asteroid to the values
     * corresponding to a moment in time given as the argument. The
     * timestamp is updated accordingly.
     *
     * @param time the new timestamp
     */
    public void updateLocationAndAngle(long time) {
        synchronized(this) {
            double deltaT = (double)(time - this.timestamp);
            this.angle = this.angle + this.angularVelocity * deltaT;
            while(this.angle < 0.0) {
                this.angle += 2.0 * Math.PI;
            }
            this.angle %= 2.0 * Math.PI;
            this.locationX = moveToRangeX(this.locationX + this.velocityX * deltaT);
            this.locationY = moveToRangeY(this.locationY + this.velocityY * deltaT);
            this.timestamp = time;
        }
    }
    
    /**
     * Makes the asteroid visible. used when a larger asteroid
     * explodes. Be sure to call <code>setLocationAndAngle</code>
     * first.
     */
    public final void makeVisible() {
        this.visible = true;
    }

    /**
     * Tells the asteroid to explode. It is up to the caller to make
     * residual asteroids visible appropriately.
     *
     * @param time the point in time when the explosion starts
     */
    public final void explode(long time) {
        int numOfVertices = this.vertices.length / 2;
        double deltaT, theta, sinTheta, cosTheta, x, y;
        synchronized(this) {
            deltaT = (double)(time - this.timestamp);
            theta = this.angle + this.angularVelocity * deltaT;
        }
        sinTheta = Math.sin(theta);
        cosTheta = Math.cos(theta);
        AffineTransform at = new AffineTransform();
        at.setTransform(cosTheta,
                        sinTheta,
                        -sinTheta,
                        cosTheta,
                        0.0,
                        0.0);
        float[] translatedVertices = new float[this.vertices.length];
        at.transform(this.vertices,
                     0,
                     translatedVertices,
                     0,
                     numOfVertices);
        // recycling the array
        this.centers = this.vertices;
        int j;
        for(int i = 0; i < numOfVertices - 1; i++) {
            j = i + 1;
            if(j == numOfVertices) {
                j = 0;
            }
            centers[i * 2] = (translatedVertices[i * 2]
                              + translatedVertices[j * 2]) / 2.0f;
            centers[i * 2 + 1] = (translatedVertices[i * 2 + 1]
                                  + translatedVertices[j * 2 + 1]) / 2.0f;
        }
        this.vertices = translatedVertices;
        synchronized(this) {
            this.locationX = moveToRangeX(this.locationX + this.velocityX * deltaT);
            this.locationY = moveToRangeY(this.locationY + this.velocityY * deltaT);
            this.timestamp = time;
            this.angle = 0.0;
            this.visible = false;
            this.exploding = true;
        }
    }

    /**
     * Tells the asteroid to explode withou debris effects. It is up to
     * the caller to make residual asteroids visible appropriately.
     */
    public void explodeOnServer() {
        this.visible = false;
    }

    /**
     * Indicates whether the asteroid is visible.
     * @return a boolean indicating visibility
     */
    public final boolean isVisible() {
        return this.visible;
    }

    /**
     * Registers a bullet hit to the asteroid and checks whether it
     * explodes.
     *
     * @return <code>true</code> if the hit causes the asteroid to explode
     *         and <code>false</code> otherwise
     */
    public boolean registerHit() {
        this.hitsToExplosion--;
        if(this.hitsToExplosion == 0) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Indicates the size of the asteroid.
     * @return the size of the asteroid
     */
    public final int getSize() {
        return this.size;
    }

    /**
     * Returns the outline color of the asteroid.
     * @return the outline color of the asteroid
     */
    public final Color getOutlineColor() {
        return SpatialObject.COLOR_TABLE[this.outlineColor];
    }

    /**
     * Returns the x coordinate of the location of the asteroid
     * at <code>time</code>.
     *
     * @param time the point in time for which the x coordinate is requested
     * @return the x coordinate of the location of the asteroid at
     *         <code>time</code>
     */
    public final double getLocationX(long time) {
        synchronized(this) {
            return moveToRangeX(this.locationX
                                + this.velocityX
                                * (double)(time - this.timestamp));
        }
    }

    /**
     * Returns the y coordinate of the location of the asteroid
     * at <code>time</code>.
     *
     * @param time the point in time for which the y coordinate is requested
     * @return the y coordinate of the location of the asteroid at
     *         <code>time</code>
     */
    public final double getLocationY(long time) {
        synchronized(this) {
            return moveToRangeY(this.locationY
                                + this.velocityY
                                * (double)(time - this.timestamp));
        }
    }

    /**
     * Returns the points awarded for blowing up this asteroid.
     * @return the points awarded for blowing up this asteroid
     */
    public final int getPoints() {
        return this.points;
    }

    /**
     * Writes the state of this asteroid to a
     * <code>ServerModelConnector</code>.
     *
     * @param thisIndex the index of this asteroid
     * @param connector the target <code>ServerModelConnector</code>
     */
    public final void writeToConnector(int thisIndex,
                                       ServerModelConnector connector)
    {
        synchronized(this) {
            try {
                connector.writeAsteroidLocationAndAngle(thisIndex,
                                                        this.locationX,
                                                        this.locationY,
                                                        this.angle,
                                                        this.timestamp);
            } catch(IOException ioe) {
                Runner.fatalError("Error in network connection!");
            }
        }
    }

    /**
     * Serializes the asteroid for sending to the clients.
     *
     * @param out the target stream
     */
    private void writeObject(java.io.ObjectOutputStream out)
        throws IOException
    {
        out.writeLong(this.timestamp);
        out.writeDouble(this.locationX);
        out.writeDouble(this.locationY);
        out.writeDouble(this.velocityX);
        out.writeDouble(this.velocityY);
        out.writeDouble(this.angularVelocity);
        out.writeInt(this.size);
        out.writeBoolean(this.visible);
        out.writeDouble(this.angle);
        out.writeObject(this.vertices);
    }

    /**
     * Deserializes the asteroid on the client side.
     *
     * @param in the source stream
     */
    private void readObject(java.io.ObjectInputStream in)
        throws IOException, ClassNotFoundException
    {
        this.timestamp = in.readLong();
        this.locationX = in.readDouble();
        this.locationY = in.readDouble();
        this.velocityX = in.readDouble();
        this.velocityY = in.readDouble();
        this.angularVelocity = in.readDouble();
        this.size = in.readInt();
        this.visible = in.readBoolean();
        this.angle = in.readDouble();
        this.vertices = (float[])in.readObject();
        int numOfVertices;
        switch(this.size) {
            case SMALL:
                numOfVertices = 7;
                break;
            case MEDIUM:
                numOfVertices = 8;
                break;
            default:
                numOfVertices = 9;
        }
        this.geometry = new GeneralPath(GeneralPath.WIND_NON_ZERO,
                                        numOfVertices + 2);
        float xCoord, yCoord;
        for(int i = 0; i < numOfVertices; i++) {
            xCoord = this.vertices[i * 2];
            yCoord = this.vertices[i * 2 + 1];
            if(i == 0) {
                this.geometry.moveTo(xCoord, yCoord);
            } else {
                this.geometry.lineTo(xCoord, yCoord);
            }
        }
        this.geometry.closePath();
        this.transform = new AffineTransform();
        this.exploding = false;
    }
}
