/*
 * Copyright 2002 Yrj Kari-Koskinen and Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.net.*;
import java.io.*;

import javax.swing.JOptionPane;

/**
 * The server-side interface to network. ServerModelConnector acts as
 * both the reader and the writer interface. The reader interface is run
 * in an independent thread. The methods which write to ObjectOutputStream
 * are synchronized and named as <code>writeFoo()</code>.
 *
 * @author Henri Sivonen, hsivonen@iki.fi
 * @author Yrj Kari-Koskinen, yrjo.kari-koskinen@hut.fi
 */
public final class ServerModelConnector implements Runnable {

    /* **************** *
     *                  *
     * Static variables *
     *                  *
     * **************** */

    /** server listens on this port */
    final static int SERVER_PORT = 9999;


    /** opcodes for the network protocol */
    final static int 
        // special opCodes
        OC_EOF = -1,
        OC_SERVER_CLOSE = 0,
        OC_CLIENT_CLOSE = 1,
        OC_GAME_OVER = 2,

        // 2x: server -> client 
        OC_ARENA = 20,
        OC_INIT_BULLET = 21,
        OC_BLOW_UP_ASTEROID = 22,
        OC_INVALIDATE_BULLET = 23,
        OC_ASTEROID_LOCATION_AND_ANGLE = 24,
        OC_BLOW_UP_SHIP = 25,
        OC_SCORE = 26,
        OC_SHIPS_LEFT = 27,
        OC_PLAYER_NAME = 28,

        // 4x: server <- client 
        OC_BULLET_FIRED = 40,

        // 6x: server <-> client 
        OC_SHIP_STATE = 60;



    /* ******************** *
     *                      *
     * Non-static variables *
     *                      *
     * ******************** */
    
    /** the socket this connector uses */
    private Socket s;

    /** the ObjectOutputStream for this connector */
    private ObjectOutputStream oos;

    /** the ObjectInputStream for this connector */
    private ObjectInputStream ois;

    /** the other ServerModelConnector */
    private ServerModelConnector otherConnector;

    /** the server's arena */
    private Arena arena;
    
    /**
     * The index of the ship owned by the client to which this connector
     * is connected. The ship indices are determined in the 
     * <code>ClientModelConnector</code>'s
     * <code>SERVER/CLIENT_SHIP_INDEX</code> constants.
     */
    private int shipIndex;

    /** 
     * The players' name of the client to which 
     * this connector is connected 
     */
    private String playerName;


    /**
     * The constructor initializes streams from the given socket.
     *
     * @param s the socket which will be used for network communication
     * @param shipIndex the index of the ship owned by the client to 
     * which this connector is connected.  The ship indices are determined 
     * in the <code>ClientModelConnector</code>'s
     * <code>SERVER/CLIENT_SHIP_INDEX</code> constants.
     */
    private ServerModelConnector(Socket s, int shipIndex) 
        throws IOException {

        this.shipIndex = shipIndex;
        this.s = s;
        this.otherConnector = null;

        this.playerName = null; 

        // initialize the streams
        this.oos = new ObjectOutputStream(s.getOutputStream());
        this.ois = new ObjectInputStream(s.getInputStream());
    }


    /**
     * This method should be called to initialize the two connectors.
     * At first it starts to listen for connections. The local client thread 
     * is also launched here so that it connects to the right connector. 
     * The two connector objects are created when the clients connect to the 
     * server. The opcode listener threads for the server connectors are 
     * also launched here.
     *
     * @param playerName this local client player's name
     *
     * @return the array containing the two connectors
     */
    public static ServerModelConnector[] 
        createServerModelConnectors(String playerName) 
        throws IOException {

        ServerModelConnector[] connectors = new ServerModelConnector[2];

        // start listening
        ServerSocket server = 
            new ServerSocket(SERVER_PORT);

        // build the local client
        Client c = new Client("localhost", true, playerName);
        new Thread(c, "Client").start();

        // take the first client
        connectors[0] = 
            new ServerModelConnector(server.accept(), 0);
        // the first client is local, so no ping is performed
        connectors[0].readPlayerName();
        // launch the opcode listener thread
        new Thread(connectors[0], "local smc").start();            

        // take the second client
        connectors[1] =
            new ServerModelConnector(server.accept(), 1);
        connectors[1].echoPing();
        connectors[1].readPlayerName();
        new Thread(connectors[1], "non-local smc").start();

        // we don't need the ServerSocket anymore
        server.close();

        // save cross references from one connector to the another
        connectors[0].otherConnector = connectors[1];
        connectors[1].otherConnector = connectors[0];
        // it's nice to have this static method, because now we can
        // fiddle with the private fields freely here :-D

        connectors[0].writePlayerName(connectors[1].shipIndex,
                                      connectors[1].playerName);
        connectors[0].flush();
        connectors[1].writePlayerName(connectors[0].shipIndex,
                                      connectors[0].playerName);
        connectors[1].flush();

        // now we need to show the local client game frame screen
        c.setVisible();

        return connectors;
    }


    /**
     * JVM runs this method when start() is called for this thread.
     * Here we start to listen for the opcodes.
     */
    public void run() {
        boolean doLoop = true;

        try {
            
            int opCode = 0;
            while(doLoop) {
                // at first we get the opCode, which is one byte long
                opCode = ois.read();
                
                switch(opCode) {

                case OC_GAME_OVER:
                case OC_SERVER_CLOSE:
                case OC_CLIENT_CLOSE:
                    internalClose(opCode);
                    doLoop = false;
                    break;

                case OC_BULLET_FIRED:
                    readBulletFired();
                    break;

                case OC_SHIP_STATE:
                    readShipState();
                    break;

                case OC_EOF:
                    // end of network stream
                    throw new IOException("Network connection ended");

                default:
                    // invalid opCode
                    throw new IOException("Error in network connection");
                }
            }
        } catch(IOException ioe) {
            Runner.fatalError("Error in network connection!");
        }
    }    


    /**
     * Closes the output stream, if it's not already closed and calls the
     * other connector to close its streams too. Finally closes also the 
     * input stream.  If the close procedure is originated from the local
     * client, an <code>exit()</code> is also done. This method is called 
     * <b>only</b> from the opcode listener loop, when it receives a 
     * close opcode.
     *
     * @param opCode the opCode which initiated the close procedure 
     * 
     * @throws IOException if the network fails to run the operations
     */
    private synchronized void internalClose(int opCode) throws IOException {
        
        if (this.oos != null) {
            ObjectOutputStream tempOos = this.oos;
            this.oos = null;

            tempOos.write(opCode);
            tempOos.flush();
            tempOos.close();

            otherConnector.close(opCode);
        }

        this.ois.close();
        this.s.close();

        // if both smcs have done the close AND it's a server originated
        // close procedure, do an exit()
        if (opCode == OC_SERVER_CLOSE && otherConnector.oos == null) {
            System.exit(0);
        }
    }


    /**
     * Closes the output stream, if it's not already closed and calls the
     * other connector to close its streams too.
     *
     * @param opCode the opCode which initiated the close procedure 
     * 
     * @throws IOException if the network fails to run the operations
     */
    private synchronized void close(int opCode) throws IOException {

        if (this.oos != null) {
            ObjectOutputStream tempOos = oos;
            this.oos = null;
            
            tempOos.write(opCode);
            tempOos.flush();
            // output stream should not be closed, because input stream
            // does not like that (weird, isn't it?)
            //tempOos.close();
            
            otherConnector.close(opCode);
        }
    }


    /**
     * A public method for initiating the game over procedure from outside this
     * connector
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized void writeGameOver() throws IOException {
        this.close(OC_GAME_OVER);
    }


    /**
     * The client sends us the clientTime, and we send the time difference back
     * The echoPing() method should be called <b>only</b> from the 
     * <code>createServerModelConnectors()</code> factory method.
     *
     * @throws IOException if the network fails to run the operations
     */
    private synchronized void echoPing() throws IOException {

        long clientTime = 0;
        for(int i = 0; i < ClientModelConnector.PING_COUNT; i++) {
            clientTime = ois.readLong();
            // send the difference = serverTime - clientTime
            oos.writeLong(System.currentTimeMillis() - clientTime);
            oos.flush();
        }
    }


    /**
     * Flushes the network output stream
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized void flush() throws IOException{
        if (this.oos != null) {
            this.oos.flush();
        }
    }
    

    /**
     * Tells the client to blow up asteroid number <code>index</code>
     *
     * @param index the index of the asteroid to blow up
     * @param locationX the x coordinate of the explosion location
     * @param locationY the y coordinate of the explosion location
     * @param timestamp the moment of explosion in game (server) time
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized void writeBlowUpAsteroid(int index,
                                                 double locationX,
                                                 double locationY,
                                                 long timestamp) 
        throws IOException {

        if (this.oos != null) {
            oos.write(OC_BLOW_UP_ASTEROID);
        
            oos.writeInt(index);
            oos.writeDouble(locationX);
            oos.writeDouble(locationY);
            
            oos.writeLong(timestamp);
        }
    }


    /**
     * Tells the client that the ship at <code>shipIndex</code> has
     * blown up.
     *
     * @param shipIndex the index of the space ship
     * @param newAppears the moment in game time when a new ship should appear
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized void writeBlowUpShip(int shipIndex,
                                             long newAppears)
        throws IOException
    {

        if (this.oos != null) {
            oos.write(OC_BLOW_UP_SHIP);
            
            oos.writeInt(shipIndex);
            oos.writeLong(newAppears);
        }
    }

    /**
     * Tells the client the number of ships a player has left.
     *
     * @param shipIndex the index of the space ship
     * @param shipsLeft the number of ships left
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized final void writeShipsLeft(int shipIndex,
                                                  int shipsLeft)
        throws IOException
    {
        if (this.oos != null) {
            oos.write(OC_SHIPS_LEFT);

            oos.writeInt(shipIndex);
            oos.writeInt(shipsLeft);
        }
    }
    
    /**
     * Marks a bullet non-flying (not visible).
     *
     * @param shipIndex the index of the ship owning the bullet
     * @param bulletIndex the index of the bullet
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized final void writeInvalidateBullet(int shipIndex, 
                                                         int bulletIndex)
        throws IOException { 

        if (this.oos != null) {
            oos.write(OC_INVALIDATE_BULLET);
            
            oos.writeInt(shipIndex);
            oos.writeInt(bulletIndex);
        }
    }


    /**
     * Sends the whole Arena object to the client
     *
     * @param arena the ready-to-send arena
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized final void writeArena(Arena arena)
        throws IOException {

        this.arena = arena;
        if (this.oos != null) {
            oos.write(OC_ARENA);
            oos.writeObject(arena);
        }
    }


    /**
     * Tells the client that a bullet has been fired by the ship
     * controlled by the other client.
     *
     * @param shipIndex the index of the ship owning the bullet
     * @param bulletIndex the index of the bullet
     * @param locationX the x coordinate at which the bullet was fired
     * @param locationY the y coordinate at which the bullet was fired
     * @param velocityX the x component of the velocity of the bullet
     * @param velocityY the y component of the velocity of the bullet
     * @param lineDeltaX the x difference of the between the center of
     *                   bullet the end of the bullet line
     * @param lineDeltaY the x difference of the between the center of
     *                   bullet the end of the bullet line
     * @param timestamp the moment at which the bullet was fired
     * @param expires the moment in time at which the bullet expires
     *
     * @throws IOException if the network fails to run the operations
     */
    private synchronized final void writeInitBullet(int shipIndex,
                                                    int bulletIndex,
                                                    double locationX,
                                                    double locationY,
                                                    double velocityX,
                                                    double velocityY,
                                                    double lineDeltaX,
                                                    double lineDeltaY,
                                                    long timestamp,
                                                    long expires)
        throws IOException {

        if (this.oos != null) {
            oos.write(ServerModelConnector.OC_INIT_BULLET);
            
            oos.writeInt(shipIndex);
            oos.writeInt(bulletIndex);
            oos.writeDouble(locationX);
            oos.writeDouble(locationY);
            oos.writeDouble(velocityX);
            oos.writeDouble(velocityY);
            oos.writeDouble(lineDeltaX);
            oos.writeDouble(lineDeltaY);
            
            oos.writeLong(timestamp);
            oos.writeLong(expires);
        }
    }


    /**
     * Tells the server that a bullet has been fired by the ship
     * controlled by this client.
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readBulletFired() throws IOException {
        int bulletIndex = ois.readInt();
        double locationX = ois.readDouble();
        double locationY = ois.readDouble();
        double velocityX = ois.readDouble();
        double velocityY = ois.readDouble();
        double lineDeltaX = ois.readDouble();
        double lineDeltaY = ois.readDouble();
        
        long timestamp = ois.readLong();
        long expires = ois.readLong();
        
        // forward the information to the other client via the other connector
        if (otherConnector != null) {
            otherConnector.writeInitBullet(this.shipIndex, 
                                           bulletIndex, locationX, locationY, 
                                           velocityX, velocityY, lineDeltaX, 
                                           lineDeltaY, timestamp, expires);
            otherConnector.flush();
        }


        Bullet b;
        b = this.arena.getSpaceShips()[this.shipIndex].
            getBullets()[bulletIndex];
        b.initFromRemote(locationX,
                         locationY,
                         velocityX,
                         velocityY,
                         lineDeltaX,
                         lineDeltaY,
                         timestamp,
                         expires);
    }


    /**
     * Sends the state of the space ship to the client.
     *
     * @param shipIndex the index of the space ship
     * @param locationX the x coordinate of the space ship
     * @param locationY the y coordinate of the space ship
     * @param velocityX the x component of the velocity of the space ship
     * @param velocityY the y component of the velocity of the space ship
     * @param angle the rotation angle of the space ship
     * @param timestamp the current server (game) time
     *
     * @throws IOException if the network fails to run the operations
     */
    private synchronized final void writeShipState(int shipIndex,
                                                   double locationX,
                                                   double locationY,
                                                   double velocityX,
                                                   double velocityY,
                                                   double angle,
                                                   long timestamp) 
        throws IOException {

        // if the stream is already closed, the crazy ship state pump
        // might want to do this again...
        if (this.oos != null) {
            oos.write(OC_SHIP_STATE);
            
            oos.writeInt(shipIndex);
            oos.writeDouble(locationX);
            oos.writeDouble(locationY);
            oos.writeDouble(velocityX);
            oos.writeDouble(velocityY);
            oos.writeDouble(angle);
            
            oos.writeLong(timestamp);
        }
    }


    /**
     * Reads the state of the space ship controlled by this connector's 
     * client, forwards the information to the other connector's client
     * and updates the information to the model.
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readShipState() throws IOException {
        int shipIndex = ois.readInt();
        double locationX = ois.readDouble();
        double locationY = ois.readDouble();
        double velocityX = ois.readDouble();
        double velocityY = ois.readDouble();
        double angle = ois.readDouble();
        
        long timestamp = ois.readLong();
        
        // forward the information to the other client via the other connector
        if (otherConnector != null) {
            otherConnector.writeShipState(shipIndex, locationX, locationY, 
                                          velocityX, velocityY, angle,
                                          timestamp);
            otherConnector.flush();
        }
        this.arena.getSpaceShips()[shipIndex].setStateFromRemote(locationX,
                                                                 locationY,
                                                                 velocityX,
                                                                 velocityY,
                                                                 angle,
                                                                 timestamp);
    }


    /**
     * Tells the client the location and angle of asteroid number 
     * <code>index</code>
     *
     * @param index the index of the asteroid to blow up
     * @param locationX the x coordinate of the explosion location
     * @param locationY the y coordinate of the explosion location
     * @param angle the rotation angle of the asteroid
     * @param timestamp the moment of explosion in game (server) time
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized void writeAsteroidLocationAndAngle(int index,
                                                           double locationX,
                                                           double locationY,
                                                           double angle,
                                                           long timestamp) 
        throws IOException {

        if (this.oos != null) {
            oos.write(OC_ASTEROID_LOCATION_AND_ANGLE);
            oos.writeInt(index);
            oos.writeDouble(locationX);
            oos.writeDouble(locationY);
            oos.writeDouble(angle);
            oos.writeLong(timestamp);
        }
    }


    /**
     * Inform the client of the new scores of a player.
     *
     * @param shipIndex the index of the space ship
     * @param score the player's score
     *
     * @throws IOException if the network fails to run the operations
     */
    public synchronized void writeScore(int shipIndex, int score)
        throws IOException {

        if (this.oos != null) {
            oos.write(OC_SCORE);
            oos.writeInt(shipIndex);
            oos.writeInt(score);
        }
    }


    /**
     * Read the player name of the client connected to this connector and
     * save it temporarily to the <code>this.playerName</code> field.
     *
     * @throws IOException if the network fails to run the operations
     */
    private final void readPlayerName() throws IOException {
        try {
            this.playerName = (String) ois.readObject();
        } catch(ClassNotFoundException cnfe) {
            Runner.fatalError("Error in Player name!");
        }
    }


    /**
     * Write the given <code>playerName</code> of the given 
     * <code>shipIndex</code>.
     *
     * @param shipIndex the index of the space ship
     * @param playerName the player's name
     *
     * @throws IOException if the network fails to run the operations
     */
    private synchronized final void writePlayerName(int shipIndex,
                                                    String playerName) 
        throws IOException { 

        if (this.oos != null) {
            oos.write(OC_PLAYER_NAME);        
            oos.writeInt(shipIndex);
            oos.writeObject(playerName);
        }
    }
}
